#!/usr/bin/env python
# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import absolute_import
from __future__ import with_statement
import sys

import dataproperty
import pyparsing as pp
import six
import thutils

import tcconfig
import tcconfig.traffic_control


def parse_option():
    parser = thutils.option.ArgumentParserObject()
    parser.make(version=tcconfig.VERSION)

    group = parser.parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--device",
        help="network device name (e.g. eth0)")
    group.add_argument(
        "-f", "--config-file",
        help="""setting traffic controls from a configuration file.
        output file of the tcshow.""")

    group = parser.add_argument_group("Network Interface")
    group.add_argument(
        "--overwrite", action="store_true", default=False,
        help="overwrite existing settings")

    group = parser.add_argument_group("Traffic Control")
    group.add_argument(
        "--direction", choices=tcconfig.traffic_control.TrafficDirection.LIST,
        default=tcconfig.traffic_control.TrafficDirection.OUTGOING,
        help="""the direction of network communication that impose traffic control.
        ``incoming`` requires linux kernel version 2.6.20 or later.
        (default = ``%(default)s``)
        """)
    group.add_argument(
        "--rate", dest="bandwidth_rate",
        help="network bandwidth rate [K|M|G bps]")
    group.add_argument(
        "--delay", dest="network_latency", type=float, default=0,
        help="round trip network delay [ms] (default=%(default)s)")
    group.add_argument(
        "--delay-distro", dest="latency_distro_ms", type=float, default=0,
        help="""
        distribution of network latency becomes X +- Y [ms]
        (normal distribution), with this option.
        (X: value of --delay option, Y: value of --delay-dist option)
        network latency distribution will be uniform without this option.
        """)
    group.add_argument(
        "--loss", dest="packet_loss_rate", type=float, default=0,
        help="round trip packet loss rate [%%] (default=%(default)s)")
    group.add_argument(
        "--corrupt", dest="corruption_rate", type=float, default=0,
        help="""
        packet corruption rate [%%].
        packet corruption means single bit error at a random offset in the packet.
        (default=%(default)s)
        """)
    group.add_argument(
        "--network",
        help="IP address/network of traffic control")
    group.add_argument(
        "--port", type=int,
        help="port number of traffic control")

    return parser.parse_args()


def _parse_tc_filter_network(text):
    network_pattern = (
        pp.SkipTo("network=", include=True) +
        pp.Word(pp.alphanums + "." + "/"))

    return network_pattern.parseString(text)[-1]


def _parse_tc_filter_port(text):
    port_pattern = (
        pp.SkipTo("port=", include=True) +
        pp.Word(pp.nums))

    return port_pattern.parseString(text)[-1]


def load_tcconfig(config_file):
    from voluptuous import Schema, Required, Any, ALLOW_EXTRA

    schema = Schema({
        Required(six.text_type): {
            Any("outgoing", "incoming"): {
                six.text_type: {
                    six.text_type: six.text_type,
                },
            }
        },
    }, extra=ALLOW_EXTRA)

    return thutils.loader.JsonLoader.load(
        config_file, schema)


def get_tcconfig_command_list(config_table, is_overwrite):
    command_list = []

    for device, device_table in six.iteritems(config_table):
        if is_overwrite:
            command_list.append("tcdel --device " + device)

        for direction, direction_table in six.iteritems(device_table):
            for tc_filter, filter_table in six.iteritems(direction_table):
                if filter_table == {}:
                    continue

                option_list = [
                    "--device=" + device,
                    "--direction=" + direction,
                ] + [
                    "--%s=%s" % (k, v)
                    for k, v in six.iteritems(filter_table)
                ]

                try:
                    network = _parse_tc_filter_network(tc_filter)
                    if network != "0.0.0.0/0":
                        option_list.append("--network=" + network)
                except pp.ParseException:
                    pass

                try:
                    port = _parse_tc_filter_port(tc_filter)
                    option_list.append("--port=" + port)
                except pp.ParseException:
                    pass

                command_list.append(" ".join(["tcset"] + option_list))

    return command_list


@thutils.main.Main
def main():
    options = parse_option()

    thutils.initialize_library(__file__, options)
    thutils.common.verify_install_command(["tc"])

    subproc_wrapper = thutils.subprocwrapper.SubprocessWrapper()

    if dataproperty.is_not_empty_string(options.config_file):
        return_code = 0

        for tcconfig_command in get_tcconfig_command_list(
                load_tcconfig(options.config_file), options.overwrite):
            return_code |= subproc_wrapper.run(tcconfig_command)

        return return_code

    tc = tcconfig.traffic_control.TrafficControl(
        subproc_wrapper, options.device)
    tc.direction = options.direction
    tc.bandwidth_rate = options.bandwidth_rate
    tc.latency_ms = options.network_latency
    tc.latency_distro_ms = options.latency_distro_ms
    tc.packet_loss_rate = options.packet_loss_rate
    tc.corruption_rate = options.corruption_rate
    tc.network = options.network
    tc.port = options.port

    tc.validate()

    if options.overwrite:
        tc.delete_tc()

    tc.set_tc()

    return 0


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from __future__ import absolute_import
from __future__ import with_statement
import sys

import thutils
import tcconfig
import tcconfig.traffic_control


def parse_option():
    parser = thutils.option.ArgumentParserObject()
    parser.make(version=tcconfig.VERSION)

    group = parser.add_argument_group("Traffic Control")
    group.add_argument(
        "--device", required=True,
        help="network device name (e.g. eth0)")

    return parser.parse_args()


@thutils.main.Main
def main():
    options = parse_option()

    thutils.initialize_library(__file__, options)

    thutils.common.verify_install_command(["tc"])
    tcconfig.verify_network_interface(options.device)

    subproc_wrapper = thutils.subprocwrapper.SubprocessWrapper()
    tc = tcconfig.traffic_control.TrafficControl(
        subproc_wrapper, options.device)

    return tc.delete_tc()


if __name__ == '__main__':
    sys.exit(main())



# encoding: utf-8

'''
@author: Tsuyoshi Hombashi
'''

import pytest
import thutils


DEVICE = "eth0"


@pytest.fixture
def subproc_wrapper():
    return thutils.subprocwrapper.SubprocessWrapper()


class Test_tcshow:
    """
    Tests of in this class are inappropriate for Travis CI.
    Execute following command at the local environment  when running tests:
      python setup.py test --addopts --runxfail
    """

    @pytest.mark.xfail
    def test_normal(self, subproc_wrapper):
        command = " ".join([
            "tcset",
            "--device", DEVICE,
            "--delay", "10",
            "--delay-distro", "2",
            "--loss", "0.01",
            "--rate", "0.25M",
            "--network", "192.168.0.10",
            "--port", "8080",
        ])
        assert subproc_wrapper.run(command) == 0

        command = " ".join([
            "tcset",
            "--device", DEVICE,
            "--delay", "1",
            "--loss", "0.02",
            "--rate", "500K",
            "--direction", "incoming",
        ])
        assert subproc_wrapper.run(command) == 0

        command = " ".join([
            "tcshow",
            "--device", DEVICE,
        ])
        proc = subproc_wrapper.popen_command(command)
        stdout, _stderr = proc.communicate()
        assert thutils.loader.JsonLoader.loads(stdout) == thutils.loader.JsonLoader.loads("""{
    "eth0": {
        "outgoing": {
            "network=192.168.0.10/32, port=8080": {
                "delay": "10.0", 
                "loss": "0.01", 
                "rate": "250K", 
                "delay-distro": "2.0"
            }, 
            "network=0.0.0.0/0": {}
        }, 
        "incoming": {
            "network=0.0.0.0/0": {
                "delay": "1.0", 
                "loss": "0.02", 
                "rate": "500K"
            }
        }
    }
}
""")

        assert subproc_wrapper.run("tcdel --device " + DEVICE) == 0



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

VERSION = "0.6.2"


def verify_network_interface(device):
    import netifaces

    if device not in netifaces.interfaces():
        raise ValueError("invalid network interface: " + device)



# encoding: utf-8

'''
@author: Tsuyoshi Hombashi
'''

from __future__ import absolute_import
import itertools

import pytest
import thutils
import tcconfig
import tcconfig.traffic_control


MIN_PACKET_LOSS = 0.0000000232  # [%]


@pytest.fixture
def tc_obj():
    subproc_wrapper = thutils.subprocwrapper.SubprocessWrapper()
    return tcconfig.traffic_control.TrafficControl(subproc_wrapper, "eth0")


@pytest.mark.parametrize(["value"], [
    ["".join(opt_list)]
    for opt_list in itertools.product(
        ["0.1", "1", "2147483647"],
        [
            "k", " k", "K", " K",
            "m", " m", "M", " M",
            "g", " g", "G", " G",
        ]
    )
])
def test_TrafficControl_validate_bandwidth_rate_normal(tc_obj, value):
    tc_obj.bandwidth_rate = value
    tc_obj._TrafficControl__validate_bandwidth_rate()


@pytest.mark.parametrize(["value", "expected"], [
    ["".join(opt_list), ValueError]
    for opt_list in itertools.product(
        ["0.1", "1", "2147483647"],
        [
            "kb", "kbps", "KB",
            "mb", "mbps", "MB",
            "gb", "gbps", "GB",
        ]
    )
])
def test_TrafficControl_validate_bandwidth_rate_exception_1(
        tc_obj, value, expected):
    tc_obj.bandwidth_rate = value
    with pytest.raises(expected):
        tc_obj._TrafficControl__validate_bandwidth_rate()


@pytest.mark.parametrize(["value", "expected"], [
    ["".join(opt_list), ValueError]
    for opt_list in itertools.product(
        ["-1", "0", "0.0"],
        ["k", "K", "m", "M", "g", "G"]
    )
])
def test_TrafficControl_validate_bandwidth_rate_exception_2(
        tc_obj, value, expected):
    tc_obj.bandwidth_rate = value
    with pytest.raises(expected):
        tc_obj._TrafficControl__validate_bandwidth_rate()


class Test_TrafficControl_validate:

    @pytest.mark.parametrize(
        [
            "rate", "delay", "delay_distro", "loss",
            "corrupt", "network", "port",
        ],
        [
            opt_list
            for opt_list in itertools.product(
                [None, "", "100K", "0.5M", "0.1G"],  # rate
                [None, 0, 10000],  # delay
                [None, 0, 10000],  # delay_distro
                [None, 0, MIN_PACKET_LOSS, 99],  # loss
                [None, 0, 99],  # corrupt
                [
                    None,
                    "",
                    "192.168.0.1", "192.168.0.254",
                    "192.168.0.1/32", "192.168.0.0/24"
                ],  # network
                [None, 0, 65535],  # port
            )
        ])
    def test_normal(
            self, tc_obj, rate, delay, delay_distro, loss, corrupt, network, port):
        tc_obj.bandwidth_rate = rate
        tc_obj.latency_ms = delay
        tc_obj.latency_distro_ms = delay_distro
        tc_obj.packet_loss_rate = loss
        tc_obj.corruption_rate = corrupt
        tc_obj.network = network
        tc_obj.port = port

        tc_obj.validate()

    @pytest.mark.parametrize(["value", "expected"], [
        [{"latency_ms": -1}, ValueError],
        [{"latency_ms": 10001}, ValueError],

        [{"latency_distro_ms": -1}, ValueError],
        [{"latency_distro_ms": 10001}, ValueError],

        [{"packet_loss_rate": -0.1}, ValueError],
        [{"packet_loss_rate": 99.1}, ValueError],

        [{"curruption_rate": -0.1}, ValueError],
        [{"curruption_rate": 99.1}, ValueError],

        [{"network": "192.168.0."}, ValueError],
        [{"network": "192.168.0.256"}, ValueError],
        [{"network": "192.168.0.0/0"}, ValueError],
        [{"network": "192.168.0.0/33"}, ValueError],
        [{"network": "192.168.0.2/24"}, ValueError],
        [{"network": "192.168.0.0000/24"}, ValueError],

        [{"port": -1}, ValueError],
        [{"port": 65536}, ValueError],
    ])
    def test_exception(self, tc_obj, value, expected):
        tc_obj.bandwidth_rate = value.get("bandwidth_rate")
        tc_obj.latency_ms = value.get("latency_ms")
        tc_obj.latency_distro_ms = value.get("latency_distro_ms")
        tc_obj.packet_loss_rate = value.get("packet_loss_rate")
        tc_obj.curruption_rate = value.get("curruption_rate")
        tc_obj.network = value.get("network")
        tc_obj.port = value.get("port")

        with pytest.raises(expected):
            tc_obj.validate()
